var map;
var infowindow;

var markers = [];
var _locations = [];
var _markersUpdatedCallback;

var selectedMarker = null;
var _markerSelectionChangedCallback;

var path;

/*
// SVG custom marker//

var pinSVGHole = "M12,11.5A2.5,2.5 0 0,1 9.5,9A2.5,2.5 0 0,1 12,6.5A2.5,2.5 0 0,1 14.5,9A2.5,2.5 0 0,1 12,11.5M12,2A7,7 0 0,0 5,9C5,14.25 12,22 12,22C12,22 19,14.25 19,9A7,7 0 0,0 12,2Z";
var labelOriginHole = new google.maps.Point(12,15);

var pinSVGFilled = "M 12,2 C 8.1340068,2 5,5.1340068 5,9 c 0,5.25 7,13 7,13 0,0 7,-7.75 7,-13 0,-3.8659932 -3.134007,-7 -7,-7 z";
var labelOriginFilled =  new google.maps.Point(12,9);

var pinColor = "#00FF00";

var markerSelectedIcon = {  // https://developers.google.com/maps/documentation/javascript/reference/marker#MarkerLabel
    path: pinSVGHole,
    anchor: new google.maps.Point(12,17),
    fillOpacity: 1,
    fillColor: pinColor,
    strokeWeight: 2,
    strokeColor: "white",
    scale: 2,
    labelOrigin: labelOriginHole
};
*/
// PNG custom marker//
var markerSelectedIcon = 'http://maps.google.com/mapfiles/kml/paddle/ylw-blank.png'; //see http://kml4earth.appspot.com/icons.html

var markerLabels = "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΥΩαβγδεζηθικλμνξοπρστυφχψω0123456789";

//--- MAP ---

function initMap(viewLocation, viewZoom, markersUpdatedCallback, markerSelectionChangedCallback) {
    _markersUpdatedCallback = markersUpdatedCallback;
    _markerSelectionChangedCallback = markerSelectionChangedCallback;

    var mapOptions = {
        zoom: viewZoom,
        center: new google.maps.LatLng(viewLocation[0], viewLocation[1]),
        logoPassive: false,
        draggingCursor: 'hand',
        disableDefaultUI: false,
        draggableCursor: 'crosshair',
        mapTypeId: google.maps.MapTypeId.SATELLITE,
        mapTypeControlOptions: {
            style: google.maps.MapTypeControlStyle.DROPDOWN_MENU
        }
    };

    map = new google.maps.Map(document.getElementById("map_canvas"), mapOptions);

    google.maps.event.addListener(map, 'click', function(event) {
        addMarker(event.latLng);
    });

    onMarkersUpdated([]);
}

//--- MARKERS ---

function markerIndex(marker) {
    var markerCount = markers.length;
    for (var i = 0; i < markerCount; i++)
        if (markers[i] == marker)
            return i;

    return -1;
}

function addMarker(location) {
    var marker = new google.maps.Marker({
            position: location,
            draggable: false, /*true*/
            title: shortLocation(location),
            icon: '',
            map: map,
            label: markerLabels[markers.length] //String.fromCharCode('Α'.charCodeAt(0) + markers.length)
        });
    markers.push(marker);

    _locations.push(location);

    // Add mouseover handler: //NOT USED, TOO JUMPY
    /*
    google.maps.event.addListener(marker, 'mouseover', function () {
        selectMarker(marker); //this also calls showInfoWindow(marker)
    });
    */
    
    //Add click handler:
    google.maps.event.addListener(marker, 'click', function () {
        //selectMarker(null); //this also calls showInfoWindow(null)
        selectMarker(marker); //this also calls showInfoWindow(marker)
    });

    //Add drag handler
    /*
    google.maps.event.addListener(marker, 'drag', function() {
        onMarkersUpdated();
        selectMarker(marker);
    });
    */

    selectMarker(marker); //this will also call showInfoWindow(marker)
    onMarkersUpdated(); //this will result in elevation API call, will return  asynchronously and update plot
}

function deleteLastMarker() {
    if (markers.length != 0) {
        var marker = markers.pop();
        marker.setMap(null);
        _locations.pop();

        selectMarker(null);
        onMarkersUpdated();
    }
}

function deleteAllMarkers() {
    if (markers) {
        var markerCount = markers.length;
        for (var i = markerCount-1; i >=0; i--) {
            markers[i].setMap(null);
            _locations.pop();
        }
        markers.length = 0;
    }

        selectMarker(null);
        onMarkersUpdated();
}

function onMarkersUpdated() {
    showInfoWindow(null); //hide info window if open
    makePath(_locations); //if _locations==[] it will clear the current path, if any

    if (_markersUpdatedCallback != null)
        _markersUpdatedCallback(_locations);
}

function onMarkerSelectionChanged(marker) {
    if (_markerSelectionChangedCallback != null)
        _markerSelectionChangedCallback(marker); 
}

function selectMarker(marker) {
    selectedMarker = marker;
    // Deselect any currently selected marker:
    /*
    for (var i=0; i < markers.length ; i++) {
        markers[i].setIcon();
    }
    */

    showInfoWindow(marker); //this will show (hide if marker==null) the infoWindow for that marker

    if (marker != null) {
        //map.panTo(marker.getPosition()); //NOT USED, TOO JUMPY
        //map.setZoom(3);

        if (!marker.getIcon()) {
            //DON'T NEED TO CHANGE MARKER ICON TO SHOW SELECTION, WE REUSE AN INFOTIP FOR THAT
            marker.setIcon(markerSelectedIcon);
            // Auto hide marker selection / Restore default marker icon: (NOTE: CUSTOM MARKERS SEEM TO HAVE WIDE SELECTION AREA, WHICH MAKES IT HARD TO ADD MARKERS ONE CLOSE TO ANOTHER - THE DEFAULT MARKER DOESN'T HAVE SUCH AN ISSUE)
            setTimeout(function() {
                marker.setIcon(); 
            }, 400);
        }
    }

    onMarkerSelectionChanged(marker); //will also notify for null (clear) selection
}

//--- INFOTIP ---

function showInfoWindow(marker) {
    if (infowindow != null)
        infowindow.close(map);
    
    if (marker == null)
        return; //thus, if we just want to close the InfoWindow we can do showInfoWindow(null)

    infowindow = new google.maps.InfoWindow({
        content: shortLocation(marker.position),
        position: marker.position
    });
    
    // Autoclose info window:
    //setTimeout(function(){ infowindow.close(map, marker); }, 1500 /*msec*/);

    infowindow.open(map, marker);
}

//--- PATH ---

function makePath(locations) {
    if (path != null)
        path.setMap(null); //can pass empty locations array to just clear the path

    if (locations.length != 0) {
        path = new google.maps.Polyline({
                path: locations,
                strokeColor: '#0000ff',
                strokeOpacity: 1.0,
                strokeWeight: 5
            });
        path.setMap(map);
    }
}

//--- HELPERS ---

function round(value, decimals) {
  return Number(Math.round(value+'e'+decimals)+'e-'+decimals);
}

function shortLocation(location) {
    return localizedNum(round(location.lat(), 3)) + ", " + localizedNum(round(location.lng(), 3));
}
