var Feature = ol.Feature; //import Feature from 'ol/Feature.js';
//var Map = ol.Map; //import Map from 'ol/Map.js'; //commented-out, use ol.Map in code instead of Map, internal JS Map object causes conflicts in newer browsers
var View = ol.View; //import View from 'ol/View.js';
var Point = ol.geom.Point; //import Point from 'ol/geom/Point.js';

var Tile = ol.layer.Tile;
var Vector = ol.layer.Vector;
var VectorTile = ol.layer.VectorTile; //import {Tile as TileLayer, Vector as VectorLayer, VectorTile As VectorTileLayer} from 'ol/layer.js';

var TileLayer = Tile;
var VectorLayer = Vector;
//var VectorTileLayer = VectorTile;

//var OSM = ol.source.OSM; //import OSM from 'ol/source/OSM.js';
var XYZ = ol.source.XYZ; //import XYZ from 'ol/source/XYZ.js';
//var BingMaps = ol.source.BingMaps; //import BingMaps from 'ol/source/BingMaps.js';
var VectorSource = ol.source.Vector; //import VectorSource from 'ol/source/Vector.js';
//var VectorTileSource = ol.source.VectorTile; //import VectorTile from 'ol/source/VectorTile.js';

var Polyline = ol.format.Polyline; //import Polyline from 'ol/format/Polyline.js';
//var GeoJSON = ol.format.GeoJSON; //import GeoJSON from 'ol/format/GeoJSON.js';

var Circle = ol.style.Circle;
var Fill = ol.style.Fill;
var Icon = ol.style.Icon;
var Stroke = ol.style.Stroke;
var Style = ol.style.Style; //import {Circle as CircleStyle, Fill, Icon, Stroke, Style} from 'ol/style.js';

var CircleStyle = Circle;

//------------------------------

var map;
var mapLayer_ESRI, /*mapLayer_OSM,*/ /*mapLayer_Bing,*/ markersLayer;
var infowindow;

var markers = [];
var _locations = [];
var _markersUpdatedCallback;

var selectedMarker = null;
var _markerSelectionChangedCallback;

var path;

var markerLabels = "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΥΩαβγδεζηθικλμνξοπρστυφχψω0123456789";

//--- MAP ---

function initMap(mapElement, viewExtent, viewLocation, viewZoom, markersUpdatedCallback, markerSelectionChangedCallback) { //Note: viewExtend only used for the zoomToExtent button 
    _markersUpdatedCallback = markersUpdatedCallback;
    _markerSelectionChangedCallback = markerSelectionChangedCallback;

    //--- Styles ---//

    var styles = {
        'route': new Style({
            stroke: new Stroke({
                width: 6,
                color: [237, 10, 0, 0.7],
                //lineDash: [.1, 5]
            })
        }),

        'icon': new Style({
            image: new Icon({
                anchor: [0.5, 1],
                src: 'data/icon.png'
            })
        }),

        'marker': new Style({
            //label: "${name}",
            image: new CircleStyle({
                radius: 7,
                fill: new Fill({
                    color: [50, 50, 50, 0.7]
                }),
                stroke: new Stroke({
                    color: [255, 255, 255, 1],
                    width: 2
                })
            }),
            text: new ol.style.Text({
                font: 'bold 11px "Open Sans", "Arial Unicode MS", "sans-serif"',
                placement: 'point',
                textBaseline: 'middle',
                offsetX: 0.1,
                offsetY: 0.5,
                stroke: new Stroke({
                    color: [0, 0, 0, 1],
                    width: 1
                }),
                fill: new ol.style.Fill({
                    color: 'white'
                })
            })
        })
    };

    //--- Map background ---//

    mapLayer_ESRI = new TileLayer({
        source: new XYZ({
            attributions: [
                'Τροφοδοσία από Esri<br />',
                'Χάρτης: Esri, DigitalGlobe, GeoEye, Earthstar Geographics, CNES/Airbus DS, USDA, USGS, AeroGRID, IGN, and the GIS User Community<br />',
                'Εικονίδια: <a href="http://www.freepik.com/" target="_blank">Freepik</a>, <a href="https://www.flaticon.com/authors/vectors-market" target="_blank">Vectors Market</a> από <a href="https://www.flaticon.com/" target="_blank">flaticon</a>'
            ],
            //attributionsCollapsible: false,
            url: 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
            maxZoom: 23
        })
    });

    /*
    mapLayer_OSM = new TileLayer({
        source: new OSM({
            //transition: 0,
            attributions: [
                'Χάρτης: <a href="https://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a> contributors μέσω <a href="https://www.openlayers.org" target="_blank">OpenLayers</a><br />',
                'Εικονίδια: <a href="http://www.freepik.com/" target="_blank">Freepik</a>, <a href="https://www.flaticon.com/authors/vectors-market" target="_blank">Vectors Market</a> από <a href="https://www.flaticon.com/" target="_blank">flaticon</a>'
            ],
            //attributionsCollapsible: false,

        }),
        //opacity: 0.5
    });
    */

    /*
    mapLayer_BingMaps = new TileLayer({
        source: new BingMaps({
                    imagerySet: 'AerialWithLabels',
                    key: 'Your Bing Maps Key from http://www.bingmapsportal.com/ here'
                })
    });
    */

    //--- Path ---//

    path = new ol.geom.LineString([]);

    routeLayer =  new ol.layer.Vector({
        source: new ol.source.Vector({
            features: [
                new ol.Feature({ type: 'route', geometry: path })
            ]
        }),
        style: function(feature) {
            return styles[feature.get('type')];
        },
        //updateWhileAnimating: true
    });

    //--- Markers ---//

    markersLayer = new VectorLayer({
        source: new VectorSource({
            features: []
        }),
        style: function(feature) {
            var style = styles[feature.get('type')];
            //style.getImage().setScale(1000/map.getSize()[0]);
            style.getText().setText(feature.get('label'));
            return style;
        }
    });

    //--- View ---//

    var theView = new View({
        //projection: 'EPSG:3857',
        enableRotation: false,
        minZoom: 2,
        maxZoom: 17,
        zoom: viewZoom,
        center: ol.proj.fromLonLat(viewLocation) //ol.proj.transform([9.6, 42], 'EPSG:4326', 'EPSG:3857');
                                                 //equivalent to "fromLonLat" (inverse is "toLonLat")
    });

    //--- Map control ---//

    var zoomToExtentImage = document.createElement('span');
    zoomToExtentImage.innerHTML = '<img src="images/home.png" width="100%" height="100%">';

    map = new ol.Map({ //don't do new Map(...) here, uses some internal JS Map object in newer browsers
        target: mapElement,
        loadTilesWhileAnimating: true,
        view: theView,

        controls: /*ol.control.defaults().extend(*/
            ((viewExtent != null)? [
                new ol.control.ZoomToExtent({
                  label: zoomToExtentImage,
                  tipLabel: "Επαναφορά",
                  extent: viewExtent
                })
            ]:[])
            .concat([
                new ol.control.Zoom({
                    zoomInTipLabel: "Μεγέθυνση χάρτη",
                    zoomOutTipLabel: "Σμίκρυνση χάρτη",
                }), //using as alternative one that already exists in defaults
                new ol.control.FullScreen({
                    tipLabel: "Πλήρης οθόνη"
                }),
                new ol.control.Attribution({
                    collapsible: true,
                    collapsed: true,
                    tipLabel: "Περί",
                })
            ])
        /*)*/,

        layers: [
            mapLayer_ESRI,
            //mapLayer_OSM,
            //mapLayer_Bing,
            routeLayer,
            markersLayer
        ]
      });

    //--- map interaction ---//

    map.on('click', map_checkMarkerSelection_Click); //singleclick
    map.on('pointermove', map_checkMarkerSelection_Hover);
    map.on('pointermove', map_checkMapDrag);

    onMarkersUpdated([]);

    createInfoWindowOverlay(undefined);
}

function map_checkMarkerSelection_Click(e) {
    var found = false;
    map.forEachFeatureAtPixel(
        e.pixel,
        function(feature, layer) {
            //alert(JSON.stringify(feature));
            if (layer === markersLayer) {
                //selectMarker(null); //this also calls showInfoWindow(null)
                //alert(JSON.stringify(feature));
                selectMarker(feature); //this also calls showInfoWindow(feature)
                found = true;
                return true;
            }
        },
        {
            hitTolerance: 0
        }
    );

    if (!found)
        addMarker(ol.proj.toLonLat(e.coordinate));
}

function map_checkMarkerSelection_Hover(e) {
    var found = false;
    map.forEachFeatureAtPixel(
        e.pixel,
        function(feature, layer) {
        //alert(JSON.stringify(feature));
        if (layer === markersLayer) {
            //selectMarker(null); //this also calls showInfoWindow(null)
            //alert(JSON.stringify(feature));
            selectMarker(feature); //this also calls showInfoWindow(feature)
                found = true;
            return true;
        }
        },
        {
            hitTolerance: 0
        }
    );

    if (!found)
        selectMarker(null); //deselect automatically
}

function map_checkMapDrag(e) {
    if (e.dragging)
        map.getTargetElement().style.cursor = 'grabbing';
    else {
        var hit = map.hasFeatureAtPixel(map.getEventPixel(e.originalEvent));
        map.getTargetElement().style.cursor = hit ? 'pointer' : '';
    }
}

//--- LAYERS ---

function setLayerVisibility(layer, visible) {
    selectMarker(null); //this will also hide info popup if open
    layer.setVisible(visible);
}

function toggleLayer(layer) {
    setLayerVisibility(layer, !layer.getVisible());
}

//--- MARKERS ---

function markerIndex(marker) {
    var markerCount = markers.length;
    for (var i = 0; i < markerCount; i++)
        if (markers[i] == marker)
            return i;

    return -1;
}

function addMarker(location) {
    var marker = new Feature({
        type: 'marker',
        geometry: new Point(ol.proj.fromLonLat(location)),
        label: markerLabels[markers.length],
    });
    markers.push(marker);
    markersLayer.getSource().addFeature(marker);

    _locations.push(toGoogleMapsPoint(location));

    // Add mouseover handler: //NOT USED, TOO JUMPY
    /*
    //TODO: selectMarker(marker); //this also calls showInfoWindow(marker)
    */

    //Add drag handler:

    var translate = new ol.interaction.Translate({
        features: new ol.Collection([marker])
    });
    /*
    translate.on('translatestart', function (evt) {
      //NOP
    });
    */
    translate.on('translating', function (evt) {
        _locations[markerIndex(marker)] = toGoogleMapsPoint(ol.proj.toLonLat(evt.coordinate/*getMarkerCoordinates(marker)*/)); //update location
        onMarkersUpdated();
        selectMarker(marker);
    });
    map.addInteraction(translate);

    //------//

    selectMarker(marker); //this will also call showInfoWindow(marker)
    onMarkersUpdated(); //this will result in elevation API call, will return  asynchronously and update plot
}

function deleteLastMarker() {
    if (markers.length != 0) {
        var marker = markers.pop();
        markersLayer.getSource().removeFeature(marker);
        _locations.pop();

        selectMarker(null);
        onMarkersUpdated();
    }
}

function deleteAllMarkers() {
    if (markers) {
        var markerCount = markers.length;
        for (var i = markerCount-1; i >=0; i--) {
            markersLayer.getSource().removeFeature(markers[i]);
            _locations.pop();
        }
        markers.length = 0;
    }

    selectMarker(null);
    onMarkersUpdated();
}

function onMarkersUpdated() {
     showInfoWindow(null); //don't do seletMarker(null) here, else it won't show InfoWindow upon addition of new marker
     makePath(_locations); //if _locations==[] it will clear the current path, if any

     if (_markersUpdatedCallback != null)
        _markersUpdatedCallback(_locations);
}

function onMarkerSelectionChanged(marker) {
    if (_markerSelectionChangedCallback != null)
        _markerSelectionChangedCallback(marker); 
}

function selectMarker(marker) {
    selectedMarker = marker;
    // Deselect any currently selected marker:
    /*
    for (var i=0; i < markers.length ; i++) {
        markers[i].setIcon(); //DUMMY, NOT IMPLEMENTED
    }
    */

    showInfoWindow(marker); //this will show (hide if marker==null) the infowindow for that marker

    if (marker != null) {
        var view = map.getView();
        //view.setCenter(getMarkerCoordinates(marker)); //NOT USED, TOO JUMPY
        //view.setZoom(3);

        /*
        if (!marker.getIcon()) {
            //DON'T NEED TO CHANGE MARKER ICON TO SHOW SELECTION, WE REUSE AN INFOTIP FOR THAT
            marker.setIcon(markerSelectedIcon); //DUMMY, NOT IMPLEMENTED
            // Auto hide marker selection / Restore default marker icon: (NOTE: CUSTOM MARKERS SEEM TO HAVE WIDE SELECTION AREA, WHICH MAKES IT HARD TO ADD MARKERS ONE CLOSE TO ANOTHER - THE DEFAULT MARKER DOESN'T HAVE SUCH AN ISSUE)
            setTimeout(function() {
                marker.setIcon(); 
            }, 400);
        }
        */
    }

    onMarkerSelectionChanged(marker); //will also notify for null (clear) selection
}

//--- InfoWindow ---

function createInfoWindowOverlay(position) {
    var container = document.getElementById('popup');
    infowindow = new ol.Overlay({
        element: container,
        autoPan: true,
        autoPanAnimation: {
            duration: 250
        },
        position: position
    });
    map.addOverlay(infowindow);
}

function showInfoWindow(marker) {
    if (infowindow != null)
        infowindow.setPosition(undefined);

    if (marker == null)
        return; //thus, if we just want to close the InfoWindow we can do showInfoWindow(null)

    var markerCoordinates = getMarkerCoordinates(marker);

    //TODO: why do we need to create it every time?
    createInfoWindowOverlay(markerCoordinates);
   //infoWindow.setPosition(markerCoordinates);

    var content = document.getElementById('popup-content');
    content.innerHTML = shortLocation(ol.proj.toLonLat(markerCoordinates));

    var closer = document.getElementById('popup-closer');
    closer.onclick = function () {
        closer.blur();
        infowindow.setPosition(undefined);
        //TODO: do we need to remove that overlay we added?
        return false;
    };
    
    // Autoclose info window:
    //setTimeout(function(){ infowindow.setPosition(undefined); }, 1500 /*msec*/);
}

//--- PATH ---

function makePath(locations) { //can pass empty locations array to just clear the path
    if (path != null)
        path.setCoordinates(getMarkersCoordinates(markers));
}

//--- HELPERS ---

function round(value, decimals) {
  return Number(Math.round(value+'e'+decimals)+'e-'+decimals);
}

function shortLocation(coordinates) {
    return localizedNum(round(coordinates[1], 3)) + ", " + localizedNum(round(coordinates[0], 3));
}

//convert [lon, lat] coordinates to Google Maps point representation that elevation.js and chart.js expect
function toGoogleMapsPoint(coordinates) {
    return { lat: function(){return coordinates[1];}, lng: function(){return coordinates[0];} };
}

function getMarkerCoordinates(marker) {
    return marker.getGeometry().getCoordinates();
}

function getMarkersCoordinates(markers) {
    var coordinates = [];
    var markerCount = markers.length;
    for (var i = 0; i < markerCount; i++)
        coordinates.push(getMarkerCoordinates(markers[i]));
    return coordinates;
}
